from jax import random
import jax.numpy as np

def _get_fans(shape):
    receptive_field = np.prod(shape[:-2])
    if len(shape) >= 2:
        fan_in, fan_out = shape[-2], shape[-1]
    elif len(shape) == 1:
        fan_in, fan_out = shape[0]
    else:
        fan_in, fan_out = 1.
    fan_in *= receptive_field
    fan_out *= receptive_field
    return fan_in, fan_out

def zeros(key, shape, dtype=np.float32): np.zeros(shape, dtype)
def ones(key, shape, dtype=np.float32): np.ones(shape, dtype)

def uniform(mean=0.):
    def init(key, shape, lim=1., dtype=np.float32):
        return random.uniform(key, shape, dtype, minval=-lim, maxval=lim)
    return init

def normal(mean=0.):
    def init(key, shape, stddev=1., dtype=np.float32):
        return random.normal(key, shape, dtype) * stddev + mean
    return init


def variance_scaling(scale, mode, distribution):
    if scale <= 0.:
        raise ValueError(f"scale must be positive float, {scale} given")
    if mode not in {"fan_in", "fan_out", "fan_avg"}:
        raise ValueError(f"Invalid mode argument: {mode}, must be either fan_in, fan_out or fan_avg")

    def init(rng, shape, dtype=np.float32):
        fan_in, fan_out = _get_fans(shape)
        gain = scale
        if mode == "fan_in":
            gain /= fan_in
        elif mode == "fan_out":
            gain /= fan_out
        elif mode == "fan_avg":
            gain /= (fan_in + fan_out) / 2
        if distribution == "truncated_normal":
            # constant from scipy.stats.truncnorm.std(a=-2, b=2, loc=0., scale=1.)
            stddev = np.sqrt(gain) / .87962566103423978
            return random.truncated_normal(rng, -2, 2, shape, dtype) * stddev
        elif distribution == "normal":
            return random.normal(rng, shape, dtype) * np.sqrt(gain)
        elif distribution == "uniform":
            lim = np.sqrt(3. * gain)
            return random.uniform(rng, shape, dtype, minval=-lim, maxval=lim)
        else:
            raise ValueError("invalid distribution for variance scaling initializer")
    return init

def glorot_uniform(scale=1.):
    return variance_scaling(scale, "fan_avg", "uniform")


def glorot_normal(scale=1.):
    return variance_scaling(scale, "fan_avg", "truncated_normal")


def leccun_uniform(scale=1.):
    return variance_scaling(scale, "fan_in", "uniform")


def leccun_normal(scale=1.):
    return variance_scaling(scale, "fan_in", "truncated_normal")


def kaiming_normal(param=0.):
    return variance_scaling(2.0 / np.sqrt(1 + param**2), "fan_in", "truncated_normal")


def kaiming_uniform(param=0.):
    return variance_scaling(2.0 / np.sqrt(1 + param**2), "fan_in", "uniform")


def orthogonal(scale=1.):
    """Initializer that generates an orthogonal matrix.
    If the shape of the tensor to initialize is two-dimensional, it is initialized
    with an orthogonal matrix obtained from the QR decomposition of a matrix of
    random numbers drawn from a normal distribution.
    If the matrix has fewer rows than columns then the output will have orthogonal
    rows. Otherwise, the output will have orthogonal columns.
    If the shape of the tensor to initialize is more than two-dimensional,
    a matrix of shape `(shape[0] * ... * shape[n - 2], shape[n - 1])`
    is initialized, where `n` is the length of the shape vector.
    The matrix is subsequently reshaped to give a tensor of the desired shape.
    Args:
        scale: multiplicative factor to apply to the orthogonal matrix
    References:
        [Saxe et al., 2014](https://openreview.net/forum?id=_wzZwKpTDF_9C)
        ([pdf](https://arxiv.org/pdf/1312.6120.pdf))
    """
    def init(rng, shape, dtype=np.float32):
        if len(shape) < 2:
            raise ValueError("The tensor to initialize must be at least two-dimensional")
        num_rows = np.prod(shape[:-1])
        num_cols = shape[-1]
        flat_shape = (max(num_rows, num_cols), max(num_rows, num_cols))

        random_mat = random.normal(rng, flat_shape, dtype)
        q, r = np.linalg.qr(random_mat)
        d = np.diag(r)
        q *= np.sign(d)
        if num_rows < num_cols:
            q = np.transpose(q)
        return scale * np.reshape(q, shape)
    return init



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2019 <João Guilherme Madeira Araújo>


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import itertools
import operator as op

import numpy as onp
from six.moves import reduce

from jax import lax
from jax import random
import jax.numpy as np
from ..activations import *
from ..initializers import *

# Each layer constructor function returns an (init_fun, apply_fun) pair, where
#   init_fun: takes an rng key and an input shape and returns an
#     (output_shape, params) pair,
#   apply_fun: takes params, inputs, and an rng key and applies the layer.


def Dense(out_dim, W_init=kaiming_uniform(), b_init=normal()):
    """Layer constructor function for a dense (fully-connected) layer."""
    def init_fun(rng, input_shape):
        output_shape = input_shape[:-1] + (out_dim,)
        k1, k2 = random.split(rng)
        W = W_init(k1, (input_shape[-1], out_dim))
        b = b_init(k2, (out_dim,), stddev=1. / np.sqrt(out_dim))
        return output_shape, (W, b)

    def apply_fun(params, inputs, **kwargs):
        W, b = params
        return np.dot(inputs, W) + b

    return init_fun, apply_fun


def GeneralConv(
    dimension_numbers,
    out_chan,
    filter_shape,
    strides=None,
    padding="VALID",
    W_init=None,
    b_init=normal(),
):
    """Layer construction function for a general convolution layer."""
    lhs_spec, rhs_spec, out_spec = dimension_numbers
    one = (1,) * len(filter_shape)
    strides = strides or one
    W_init = W_init or kaiming_uniform()

    def init_fun(rng, input_shape):
        filter_shape_iter = iter(filter_shape)
        kernel_shape = [
            out_chan
            if c == "O"
            else input_shape[lhs_spec.index("C")]
            if c == "I"
            else next(filter_shape_iter)
            for c in rhs_spec
        ]
        output_shape = lax.conv_general_shape_tuple(
            input_shape, kernel_shape, strides, padding, dimension_numbers
        )
        bias_shape = [out_chan if c == "C" else 1 for c in out_spec]
        bias_shape = tuple(itertools.dropwhile(lambda x: x == 1, bias_shape))
        k1, k2 = random.split(rng)

        b_std = 1. / np.sqrt(np.prod(kernel_shape[:-1]))
        W, b = W_init(k1, kernel_shape), b_init(k2, bias_shape, stddev=b_std)
        return output_shape, (W, b)

    def apply_fun(params, inputs, **kwargs):
        W, b = params
        return (
            lax.conv_general_dilated(
                inputs, W, strides, padding, one, one, dimension_numbers
            ) + b
        )

    return init_fun, apply_fun


Conv1D = functools.partial(GeneralConv, ("NHC", "HIO", "NHC"))
Conv = functools.partial(GeneralConv, ("NHWC", "HWIO", "NHWC"))

def DepthwiseConv2D(out_chan,
                    filter_shape,
                    strides=None,
                    padding="VALID",
                    W_init=None,
                    b_init=normal(),
                    ):
    one = (1,) * len(filter_shape)
    strides = strides or one
    W_init = W_init or kaiming_uniform()

    def init_fun(rng, input_shape):
        kernel_shape = (filter_shape[0], filter_shape[1], 1,
                        out_chan * input_shape[3])
        output_shape = lax.conv_general_shape_tuple(
            input_shape, kernel_shape, strides, padding,
            ("NHWC", "HWIO", "NHWC")
        )
        bias_shape = tuple(input_shape[0], out_chan * input_shape[3])
        k1, k2 = random.split(rng)

        b_std = (1. / np.sqrt(np.prod(kernel_shape[:-1])))
        W, b = W_init(k1, kernel_shape), b_init(k2, bias_shape, stddev=b_std)
        return output_shape, (W, b)

    def apply_fun(params, inputs, **kwargs):
        W, b = params
        return (
            lax.conv_general_dilated(
                inputs, W, strides, padding, one, one, 
                ("NHWC", "HWIO", "NHWC"), feature_group_count=inputs[1]
            ) + b
        )

    return init_fun, apply_fun



def GeneralConvTranspose(
    dimension_numbers,
    out_chan,
    filter_shape,
    strides=None,
    padding="VALID",
    W_init=None,
    b_init=normal(),
):
    """Layer construction function for a general transposed-convolution layer."""
    lhs_spec, rhs_spec, out_spec = dimension_numbers
    one = (1,) * len(filter_shape)
    strides = strides or one
    W_init = W_init or kaiming_uniform()

    def init_fun(rng, input_shape):
        filter_shape_iter = iter(filter_shape)
        kernel_shape = [
            out_chan
            if c == "O"
            else input_shape[lhs_spec.index("C")]
            if c == "I"
            else next(filter_shape_iter)
            for c in rhs_spec
        ]
        output_shape = lax.conv_transpose_shape_tuple(
            input_shape, kernel_shape, strides, padding, dimension_numbers
        )
        bias_shape = [out_chan if c == "C" else 1 for c in out_spec]
        bias_shape = tuple(itertools.dropwhile(lambda x: x == 1, bias_shape))
        k1, k2 = random.split(rng)
        b_std = (1. / np.sqrt(np.prod(kernel_shape[:-1])))
        W, b = W_init(k1, kernel_shape), b_init(k2, bias_shape, stddev=b_std)
        return output_shape, (W, b)

    def apply_fun(params, inputs, **kwargs):
        W, b = params
        return lax.conv_transpose(inputs, W, strides, padding, dimension_numbers) + b

    return init_fun, apply_fun


Conv1DTranspose = functools.partial(GeneralConvTranspose, ("NHC", "HIO", "NHC"))
ConvTranspose = functools.partial(GeneralConvTranspose, ("NHWC", "HWIO", "NHWC"))


def LSTM(out_dim, W_init=glorot_uniform(), b_init=normal()):
    def init_fun(rng, input_shape):
        k1, k2 = random.split(rng)
        cell, hidden = b_init(k1, (out_dim,)), b_init(k2, (out_dim,))

        k1, k2, k3 = random.split(k1, num=3)
        forget_W, forget_U, forget_b = (
            W_init(k1, (*input_shape[:-1], out_dim)),
            W_init(k2, (out_dim, out_dim)),
            b_init(k3, (out_dim,)),
        )

        k1, k2, k3 = random.split(k1, num=3)
        in_W, in_U, in_b = (
            W_init(k1, (*input_shape[:-1], out_dim)),
            W_init(k2, (out_dim, out_dim)),
            b_init(k3, (out_dim,)),
        )

        k1, k2, k3 = random.split(k1, num=3)
        out_W, out_U, out_b = (
            W_init(k1, (*input_shape[:-1], out_dim)),
            W_init(k2, (out_dim, out_dim)),
            b_init(k3, (out_dim,)),
        )

        k1, k2, k3 = random.split(k1, num=3)
        change_W, change_U, change_b = (
            W_init(k1, (*input_shape[:-1], out_dim)),
            W_init(k2, (out_dim, out_dim)),
            b_init(k3, (out_dim,)),
        )

        output_shape = input_shape[:-1] + (out_dim,)
        return (
            output_shape,
            (
                (cell, hidden),
                (forget_W, forget_U, forget_b),
                (in_W, in_U, in_b),
                (out_W, out_U, out_b),
                (change_W, change_U, change_b),
            ),
        )

    def apply_fun(params, inputs, **kwargs):
        (cell, hidden), (forget_W, forget_U, forget_b), (in_W, in_U, in_b), (
            out_W,
            out_U,
            out_b,
        ), (change_W, change_U, change_b) = params

        for inp in inputs:
            input_gate = sigmoid(np.dot(inp, in_W) + np.dot(hidden, in_U) + in_b)
            change_gate = np.tanh(
                np.dot(inp, change_W) + np.dot(hidden, change_U) + change_b
            )
            forget_gate = sigmoid(
                np.dot(inp, forget_W) + np.dot(hidden, forget_U) + forget_b
            )

            cell = np.multiply(change_gate, input_gate) + np.multiply(cell, forget_gate)

            output_gate = sigmoid(np.dot(inp, out_W) + np.dot(hidden, out_U) + out_b)
            output = np.multiply(output_gate, np.tanh(cell))
            hidden = output

        return output

    return init_fun, apply_fun


def BatchNorm(
    axis=(0, 1, 2),
    epsilon=1e-5,
    center=True,
    scale=True,
    beta_init=zeros,
    gamma_init=ones,
):
    """Layer construction function for a batch normalization layer."""
    _beta_init = lambda rng, shape: beta_init(rng, shape) if center else ()
    _gamma_init = lambda rng, shape: gamma_init(rng, shape) if scale else ()
    axis = (axis,) if np.isscalar(axis) else axis

    def init_fun(rng, input_shape):
        shape = tuple(d for i, d in enumerate(input_shape) if i not in axis)
        k1, k2 = random.split(rng)
        beta, gamma = _beta_init(k1, shape), _gamma_init(k2, shape)
        return input_shape, (beta, gamma)

    def apply_fun(params, x, **kwargs):
        beta, gamma = params
        # TODO(phawkins): np.expand_dims should accept an axis tuple.
        # (https://github.com/numpy/numpy/issues/12290)
        ed = tuple(None if i in axis else slice(None) for i in range(np.ndim(x)))
        beta = beta[ed]
        gamma = gamma[ed]
        mean, var = np.mean(x, axis, keepdims=True), fastvar(x, axis, keepdims=True)
        z = (x - mean) / np.sqrt(var + epsilon)
        if center and scale:
            return gamma * z + beta
        if center:
            return z + beta
        if scale:
            return gamma * z
        return z

    return init_fun, apply_fun


def elementwise(fun, **fun_kwargs):
    """Layer that applies a scalar function elementwise on its inputs."""
    init_fun = lambda rng, input_shape: (input_shape, ())
    apply_fun = lambda params, inputs, **kwargs: fun(inputs, **fun_kwargs)
    return init_fun, apply_fun


Tanh = elementwise(np.tanh)
Relu = elementwise(relu)
Exp = elementwise(np.exp)
LogSoftmax = elementwise(logsoftmax, axis=-1)
Softmax = elementwise(softmax, axis=-1)
Softplus = elementwise(softplus)
Sigmoid = elementwise(sigmoid)
Elu = elementwise(elu)
LeakyRelu = elementwise(leaky_relu)


def _pooling_layer(reducer, init_val, rescaler=None):
    def PoolingLayer(window_shape, strides=None, padding="VALID"):
        """Layer construction function for a pooling layer."""
        strides = strides or (1,) * len(window_shape)
        rescale = rescaler(window_shape, strides, padding) if rescaler else None
        dims = (1,) + window_shape + (1,)  # NHWC
        strides = (1,) + strides + (1,)

        def init_fun(rng, input_shape):
            out_shape = lax.reduce_window_shape_tuple(
                input_shape, dims, strides, padding
            )
            return out_shape, ()

        def apply_fun(params, inputs, **kwargs):
            out = lax.reduce_window(inputs, init_val, reducer, dims, strides, padding)
            return rescale(out, inputs) if rescale else out

        return init_fun, apply_fun

    return PoolingLayer


MaxPool = _pooling_layer(lax.max, -np.inf)
SumPool = _pooling_layer(lax.add, 0.0)


def _normalize_by_window_size(dims, strides, padding):
    def rescale(outputs, inputs):
        one = np.ones(inputs.shape[1:-1], dtype=inputs.dtype)
        window_sizes = lax.reduce_window(one, 0.0, lax.add, dims, strides, padding)
        return outputs / window_sizes[..., np.newaxis]

    return rescale


AvgPool = _pooling_layer(lax.add, 0.0, _normalize_by_window_size)


def Flatten():
    """Layer construction function for flattening all but the leading dim."""

    def init_fun(rng, input_shape):
        output_shape = input_shape[0], reduce(op.mul, input_shape[1:], 1)
        return output_shape, ()

    def apply_fun(params, inputs, **kwargs):
        return np.reshape(inputs, (inputs.shape[0], -1))

    return init_fun, apply_fun


Flatten = Flatten()


def Identity():
    """Layer construction function for an identity layer."""
    init_fun = lambda rng, input_shape: (input_shape, ())
    apply_fun = lambda params, inputs, **kwargs: inputs
    return init_fun, apply_fun


Identity = Identity()


def FanOut(num):
    """Layer construction function for a fan-out layer."""
    init_fun = lambda rng, input_shape: ([input_shape] * num, ())
    apply_fun = lambda params, inputs, **kwargs: [inputs] * num
    return init_fun, apply_fun


def FanInSum():
    """Layer construction function for a fan-in sum layer."""
    init_fun = lambda rng, input_shape: (input_shape[0], ())
    apply_fun = lambda params, inputs, **kwargs: sum(inputs)
    return init_fun, apply_fun


FanInSum = FanInSum()


def FanInConcat(axis=-1):
    """Layer construction function for a fan-in concatenation layer."""

    def init_fun(rng, input_shape):
        ax = axis % len(input_shape[0])
        concat_size = sum(shape[ax] for shape in input_shape)
        out_shape = input_shape[0][:ax] + (concat_size,) + input_shape[0][ax + 1 :]
        return out_shape, ()

    def apply_fun(params, inputs, **kwargs):
        return np.concatenate(inputs, axis)

    return init_fun, apply_fun


def Dropout(rate, mode="train"):
    """Layer construction function for a dropout layer, turning off
        a the weights with given rate."""

    def init_fun(rng, input_shape):
        return input_shape, ()

    def apply_fun(params, inputs, **kwargs):
        rng = kwargs.get("rng", None)
        if rng is None:
            msg = (
                "Dropout layer requires apply_fun to be called with a PRNG key "
                "argument. That is, instead of `apply_fun(params, inputs)`, call "
                "it like `apply_fun(params, inputs, key)` where `key` is a "
                "jax.random.PRNGKey value."
            )
            raise ValueError(msg)
        if mode == "train":
            keep = random.bernoulli(rng, 1 - rate, inputs.shape)
            return np.where(keep, inputs / (1. - rate), 0.)
        else:
            return inputs

    return init_fun, apply_fun


# Composing layers via combinators


def serial(*layers):
    """Combinator for composing layers in serial.

  Args:
    *layers: a sequence of layers, each an (init_fun, apply_fun) pair.

  Returns:
    A new layer, meaning an (init_fun, apply_fun) pair, representing the serial
    composition of the given sequence of layers.
  """
    nlayers = len(layers)
    init_funs, apply_funs = zip(*layers)

    def init_fun(rng, input_shape):
        params = []
        for init_fun in init_funs:
            rng, layer_rng = random.split(rng)
            input_shape, param = init_fun(layer_rng, input_shape)
            params.append(param)
        return input_shape, params

    def apply_fun(params, inputs, **kwargs):
        rng = kwargs.pop("rng", None)
        rngs = random.split(rng, nlayers) if rng is not None else (None,) * nlayers
        for fun, param, rng in zip(apply_funs, params, rngs):
            inputs = fun(param, inputs, rng=rng, **kwargs)
        return inputs

    return init_fun, apply_fun


def parallel(*layers):
    """Combinator for composing layers in parallel.

  The layer resulting from this combinator is often used with the FanOut and
  FanInSum layers.

  Args:
    *layers: a sequence of layers, each an (init_fun, apply_fun) pair.

  Returns:
    A new layer, meaning an (init_fun, apply_fun) pair, representing the
    parallel composition of the given sequence of layers. In particular, the
    returned layer takes a sequence of inputs and returns a sequence of outputs
    with the same length as the argument `layers`.
  """
    nlayers = len(layers)
    init_funs, apply_funs = zip(*layers)

    def init_fun(rng, input_shape):
        rngs = random.split(rng, nlayers)
        return zip(
            *[
                init(rng, shape)
                for init, rng, shape in zip(init_funs, rngs, input_shape)
            ]
        )

    def apply_fun(params, inputs, **kwargs):
        rng = kwargs.pop("rng", None)
        rngs = random.split(rng, nlayers) if rng is not None else (None,) * nlayers
        return [
            f(p, x, rng=r, **kwargs)
            for f, p, x, r in zip(apply_funs, params, inputs, rngs)
        ]

    return init_fun, apply_fun


def shape_dependent(make_layer):
    """Combinator to delay layer constructor pair until input shapes are known.

  Args:
    make_layer: a one-argument function that takes an input shape as an argument
      (a tuple of positive integers) and returns an (init_fun, apply_fun) pair.

  Returns:
    A new layer, meaning an (init_fun, apply_fun) pair, representing the same
    layer as returned by `make_layer` but with its construction delayed until
    input shapes are known.
  """

    def init_fun(rng, input_shape):
        return make_layer(input_shape)[0](rng, input_shape)

    def apply_fun(params, inputs, **kwargs):
        return make_layer(inputs.shape)[1](params, inputs, **kwargs)

    return init_fun, apply_fun



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Optimizers for use with JAX.

This module contains some convenient optimizer definitions, specifically
initialization and update functions, which can be used with ndarrays or
arbitrarily-nested tuple/list/dicts of ndarrays.

An optimizer is modeled as an ``(init_fun, update_fun, get_params)`` triple of
functions, where the component functions have these signatures:

::

  init_fun(params)

  Args:
    params: pytree representing the initial parameters.

  Returns:
    A pytree representing the initial optimizer state, which includes the
    initial parameters and may also include auxiliary values like initial
    momentum. The optimizer state pytree structure generally differs from that
    of `params`.

::

  update_fun(step, grads, opt_state)

  Args:
    step: integer representing the step index.
    grads: a pytree with the same structure as `get_params(opt_state)`
      representing the gradients to be used in updating the optimizer state.
    opt_state: a pytree representing the optimizer state to be updated.

  Returns:
    A pytree with the same structure as the `opt_state` argument representing
    the updated optimizer state.

::

  get_params(opt_state)

  Args:
    opt_state: pytree representing an optimizer state.

  Returns:
    A pytree representing the parameters extracted from `opt_state`, such that
    the invariant `params == get_params(init_fun(params))` holds true.


Notice that an optimizer implementation has a lot of flexibility in the form of
opt_state: it just has to be a pytree of JaxTypes (so that it can be passed to
the JAX transforms defined in api.py) and it has to be consumable by update_fun
and get_params.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import namedtuple
import functools

from six.moves import reduce

import jax.numpy as np
from jax.util import partial, safe_zip, safe_map, unzip2
from jax import tree_util
from jax.tree_util import tree_map, tree_flatten, tree_unflatten, register_pytree_node

map = safe_map
zip = safe_zip


# The implementation here basically works by flattening pytrees. There are two
# levels of pytrees to think about: the pytree of params, which we can think of
# as defining an "outer pytree", and a pytree produced by applying init_fun to
# each leaf of the params pytree, which we can think of as the "inner pytrees".
# Since pytrees can be flattened, that structure is isomorphic to a list of
# lists (with no further nesting).

pack = tuple
OptimizerState = namedtuple(
    "OptimizerState", ["packed_state", "tree_def", "subtree_defs"]
)
register_pytree_node(
    OptimizerState,
    lambda xs: ((xs.packed_state,), (xs.tree_def, xs.subtree_defs)),
    lambda data, xs: OptimizerState(xs[0], data[0], data[1]),
)


def optimizer(opt_maker):
    """Decorator to make an optimizer defined for arrays generalize to containers.

    With this decorator, you can write init, update, and get_params functions that
    each operate only on single arrays, and convert them to corresponding
    functions that operate on pytrees of parameters. See the optimizers defined in
    optimizers.py for examples.

    Args:
      opt_maker: a function that returns an ``(init_fun, update_fun, get_params)``
        triple of functions that might only work with ndarrays, as per

        .. code-block:: haskell

            init_fun :: ndarray -> OptStatePytree ndarray
            update_fun :: OptStatePytree ndarray -> OptStatePytree ndarray
            get_params :: OptStatePytree ndarray -> ndarray

    Returns:
      An ``(init_fun, update_fun, get_params)`` triple of functions that work on
      arbitrary pytrees, as per

      .. code-block:: haskell

            init_fun :: ParameterPytree ndarray -> OptimizerState
            update_fun :: OptimizerState -> OptimizerState
            get_params :: OptimizerState -> ParameterPytree ndarray

      The OptimizerState pytree type used by the returned functions is isomorphic
      to ``ParameterPytree (OptStatePytree ndarray)``, but may store the state
      instead as e.g. a partially-flattened data structure for performance.
    """

    @functools.wraps(opt_maker)
    def tree_opt_maker(*args, **kwargs):
        init, update, get_params = opt_maker(*args, **kwargs)

        @functools.wraps(init)
        def tree_init(x0_tree):
            x0_flat, tree = tree_flatten(x0_tree)
            initial_states = [init(x0) for x0 in x0_flat]
            states_flat, subtrees = unzip2(map(tree_flatten, initial_states))
            packed_state = pack(map(pack, states_flat))
            return OptimizerState(packed_state, tree, subtrees)

        @functools.wraps(update)
        def tree_update(i, grad_tree, opt_state):
            packed_state, tree, subtrees = opt_state
            grad_flat, tree2 = tree_flatten(grad_tree)
            if tree2 != tree:
                msg = (
                    "optimizer update function was passed a gradient tree that did "
                    "not match the parameter tree structure with which it was "
                    "initialized: parameter tree {} and grad tree {}."
                )
                raise TypeError(msg.format(tree, tree2))
            states = map(tree_unflatten, subtrees, packed_state)
            new_states = map(partial(update, i), grad_flat, states)
            new_states_flat, subtrees2 = unzip2(map(tree_flatten, new_states))
            for subtree, subtree2 in zip(subtrees, subtrees2):
                if subtree2 != subtree:
                    msg = (
                        "optimizer update function produced an output structure that "
                        "did not match its input structure: input {} and output {}."
                    )
                    raise TypeError(msg.format(subtree, subtree2))
            new_packed_state = pack(map(pack, new_states_flat))
            return OptimizerState(new_packed_state, tree, subtrees)

        @functools.wraps(get_params)
        def tree_get_params(opt_state):
            packed_state, tree, subtrees = opt_state
            states = map(tree_unflatten, subtrees, packed_state)
            params = map(get_params, states)
            return tree_unflatten(tree, params)

        return tree_init, tree_update, tree_get_params

    return tree_opt_maker


### optimizers


@optimizer
def sgd(step_size):
    """Construct optimizer triple for stochastic gradient descent.

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        return x0

    def update(i, g, x):
        return x - step_size(i) * g

    def get_params(x):
        return x

    return init, update, get_params


@optimizer
def momentum(step_size, mass):
    """Construct optimizer triple for SGD with Nesterov momentum.

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        v0 = np.zeros_like(x0)
        return x0, v0

    def update(i, g, state):
        x, velocity = state
        velocity = mass * velocity - (1.0 - mass) * g
        x = x + step_size(i) * velocity
        return x, velocity

    def get_params(state):
        x, _ = state
        return x

    return init, update, get_params


@optimizer
def adagrad(step_size, momentum=0.9):
    """Construct optimizer triple for Adagrad.

    Adaptive Subgradient Methods for Online Learning and Stochastic Optimization:
    http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.
      momentum: optional, a positive scalar value for momentum

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        g_sq = np.zeros_like(x0)
        m = np.zeros_like(x0)
        return x0, g_sq, m

    def update(i, g, state):
        x, g_sq, m = state
        g_sq += g ** 2
        g_sq_inv_sqrt = np.where(g_sq > 0, 1.0 / np.sqrt(g_sq), 0.0)
        m = (1.0 - momentum) * (g * g_sq_inv_sqrt) + momentum * m
        x = x - step_size(i) * m
        return x, g_sq, m

    def get_params(state):
        x, _, _ = state
        return x

    return init, update, get_params


@optimizer
def rmsprop(step_size, gamma=0.9, eps=1e-8):
    """Construct optimizer triple for RMSProp.

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.
        gamma: Decay parameter.
        eps: Epsilon parameter.

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        avg_sq_grad = np.zeros_like(x0)
        return x0, avg_sq_grad

    def update(i, g, state):
        x, avg_sq_grad = state
        avg_sq_grad = avg_sq_grad * gamma + g ** 2 * (1.0 - gamma)
        x = x - step_size(i) * g / (np.sqrt(avg_sq_grad) + eps)
        return x, avg_sq_grad

    def get_params(state):
        x, _ = state
        return x

    return init, update, get_params


@optimizer
def rmsprop_momentum(step_size, gamma=0.9, eps=1e-8, momentum=0.9):
    """Construct optimizer triple for RMSProp with momentum.

    This optimizer is separate from the rmsprop optimizer because it needs to
    keep track of additional parameters.

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.
      gamma: Decay parameter.
      eps: Epsilon parameter.
      momentum: Momentum parameter.

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        avg_sq_grad = np.zeros_like(x0)
        mom = np.zeros_like(x0)
        return x0, avg_sq_grad, mom

    def update(i, g, state):
        x, avg_sq_grad, mom = state
        avg_sq_grad = avg_sq_grad * gamma + g ** 2 * (1.0 - gamma)
        mom = momentum * mom + step_size(i) * g / np.sqrt(avg_sq_grad + eps)
        x = x - mom
        return x, avg_sq_grad, mom

    def get_params(state):
        x, _, _ = state
        return x

    return init, update, get_params


@optimizer
def adam(step_size, b1=0.9, b2=0.999, eps=1e-8):
    """Construct optimizer triple for Adam.

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.
      b1: optional, a positive scalar value for beta_1, the exponential decay rate
        for the first moment estimates (default 0.9).
      b2: optional, a positive scalar value for beta_2, the exponential decay rate
        for the second moment estimates (default 0.999).
      eps: optional, a positive scalar value for epsilon, a small constant for
        numerical stability (default 1e-8).

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        m0 = np.zeros_like(x0)
        v0 = np.zeros_like(x0)
        return x0, m0, v0

    def update(i, g, state):
        x, m, v = state
        m = (1 - b1) * g + b1 * m  # First  moment estimate.
        v = (1 - b2) * (g ** 2) + b2 * v  # Second moment estimate.
        mhat = m / (1 - b1 ** (i + 1))  # Bias correction.
        vhat = v / (1 - b2 ** (i + 1))
        x = x - step_size(i) * mhat / (np.sqrt(vhat) + eps)
        return x, m, v

    def get_params(state):
        x, m, v = state
        return x

    return init, update, get_params


@optimizer
def sm3(step_size, momentum=0.9):
    """Construct optimizer triple for SM3.
      Memory-Efficient Adaptive Optimization for Large-Scale Learning.
      https://arxiv.org/abs/1901.11150

      Args:
        step_size: positive scalar, or a callable representing a step size schedule
          that maps the iteration index to positive scalar.
        momentum: optional, a positive scalar value for momentum

      Returns:
        An (init_fun, update_fun, get_params) triple.
      """
    step_size = make_schedule(step_size)

    def splice(seq, i, x):
        lst = list(seq)
        lst[i : i + 1] = x
        return lst

    def broadcast_into(ndim, x, axis):
        idx = splice([None] * ndim, axis, [slice(None)])
        return x[tuple(idx)]

    def init(x0):
        vs = [np.zeros(sz, dtype=x0.dtype) for sz in x0.shape]
        return x0, np.zeros_like(x0), vs

    def update(i, g, state):
        x, m, vs = state
        vs = [broadcast_into(g.ndim, v, i) for i, v in enumerate(vs)]
        accum = reduce(np.minimum, vs) + g ** 2
        accum_inv_sqrt = np.where(accum > 0, 1.0 / np.sqrt(accum), 0)
        m = (1.0 - momentum) * (g * accum_inv_sqrt) + momentum * m
        x = x - step_size(i) * m
        vs = [accum.max(splice(range(x.ndim), j, [])) for j in range(x.ndim)]
        return x, m, vs

    def get_params(state):
        x, _, _ = state
        return x

    return init, update, get_params


### learning rate schedules


def constant(step_size):
    def schedule(i):
        return step_size

    return schedule


def exponential_decay(step_size, decay_steps, decay_rate):
    def schedule(i):
        return step_size * decay_rate ** (i / decay_steps)

    return schedule


def inverse_time_decay(step_size, decay_steps, decay_rate, staircase=False):
    if staircase:

        def schedule(i):
            return step_size / (1 + decay_rate * np.floor(i / decay_steps))

    else:

        def schedule(i):
            return step_size / (1 + decay_rate * i / decay_steps)

    return schedule


def piecewise_constant(boundaries, values):
    boundaries = np.array(boundaries)
    values = np.array(values)
    if not boundaries.ndim == values.ndim == 1:
        raise ValueError("boundaries and values must be sequences")
    if not boundaries.shape[0] == values.shape[0] - 1:
        raise ValueError("boundaries length must be one longer than values length")

    def schedule(i):
        return values[np.sum(i > boundaries)]

    return schedule


def make_schedule(scalar_or_schedule):
    if callable(scalar_or_schedule):
        return scalar_or_schedule
    elif np.ndim(scalar_or_schedule) == 0:
        return constant(scalar_or_schedule)
    else:
        raise TypeError(type(scalar_or_schedule))


### utilities


def l2_norm(tree):
    """Compute the l2 norm of a pytree of arrays. Useful for weight decay."""
    leaves, _ = tree_flatten(tree)
    return np.sqrt(sum(np.vdot(x, x) for x in leaves))


def clip_grads(grad_tree, max_norm):
    """Clip gradients stored as a pytree of arrays to maximum norm `max_norm`."""
    norm = l2_norm(grad_tree)
    normalize = lambda g: np.where(norm < max_norm, g, g * (max_norm / norm))
    return tree_map(normalize, grad_tree)


### serialization utilities


class JoinPoint(object):
    """Marks the boundary between two joined (nested) pytrees."""

    def __init__(self, subtree):
        self.subtree = subtree

    # Since pytrees are containers of numpy arrays, look iterable.
    def __iter__(self):
        yield self.subtree


def unpack_optimizer_state(opt_state):
    """Converts an OptimizerState to a marked pytree.

    Converts an OptimizerState to a marked pytree with the leaves of the outer
    pytree represented as JoinPoints to avoid losing information. This function is
    intended to be useful when serializing optimizer states.

    Args:
      opt_state: An OptimizerState
    Returns:
      A pytree with JoinPoint leaves that contain a second level of pytrees.
    """
    packed_state, tree_def, subtree_defs = opt_state
    subtrees = map(tree_unflatten, subtree_defs, packed_state)
    sentinels = [JoinPoint(subtree) for subtree in subtrees]
    return tree_util.tree_unflatten(tree_def, sentinels)


def pack_optimizer_state(marked_pytree):
    """Converts a marked pytree to an OptimizerState.

    The inverse of unpack_optimizer_state. Converts a marked pytree with the
    leaves of the outer pytree represented as JoinPoints back into an
    OptimizerState. This function is intended to be useful when deserializing
    optimizer states.

    Args:
      marked_pytree: A pytree containing JoinPoint leaves that hold more pytrees.
    Returns:
      An equivalent OptimizerState to the input argument.
    """
    sentinels, tree_def = tree_flatten(marked_pytree)
    assert all(isinstance(s, JoinPoint) for s in sentinels)
    subtrees = [s.subtree for s in sentinels]
    states_flat, subtree_defs = unzip2(map(tree_flatten, subtrees))
    packed_state = pack(map(pack, states_flat))
    return OptimizerState(packed_state, tree_def, subtree_defs)



from jax import vmap
import jax.numpy as np


def squared_error(y, y_pred): return np.linalg.norm(y - y_pred)**2
def absolute_error(y, y_pred): return np.linalg.norm(y - y_pred, 1)
def hinge(y, y_pred): return np.mean(np.maximum(1. - y * y_pred, 0.), axis=-1)
def crossentropy(y, y_pred): return -np.sum(y * np.log(y_pred + np.finfo(float).eps) + (1. - y) * np.log(1. - y_pred + np.finfo(float).eps))

def create_loss(net, loss):
    def loss_fn(params, inputs, targets):
        preds = net(params, inputs)
        return loss(targets, preds)
    return loss_fn

def batch_loss(net):
    def create_batched_loss(loss):
        def loss_fn(params, inputs, targets):
            preds = net(params, inputs)
            losses = vmap((loss))(targets, preds)
            return np.mean(losses)
        return loss_fn
    return create_batched_loss



import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="fastax",
    version="0.1.1",
    description="A Jax based neural network library for research",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/joaogui1/fastax",
    author="João Guilherme Madeira Araújo",
    author_email="joaoguilhermearujo@gmail.com",
    license="Apache-2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["jax", "jaxlib", "numpy"],
)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import jax.numpy as np
from jax.scipy.special import logsumexp, erf


def sigmoid(x):
    return 1.0 / (1.0 + np.exp(-x))


def relu(x, alpha=1.0):
    return np.clip(alpha * x, a_min=0.0)


def leaky_relu(x, alpha=1.0):
    return np.where(x >= 0, alpha * x, 0.01 * x)


def elu(x, alpha=1.0):
    return np.where(x > 0, x, alpha * np.exp(x) - 1)


def gelu(x):
    return x * erf(x)


def softplus(x):
    return np.logaddexp(x, 0.0)


def softmax(x, axis=-1):
    """Apply softmax to a vector of logits x along an axis"""
    numerator = np.exp(x - x.max(axis, keepdims=True))
    return numerator / numerator.sum(axis, keepdims=True)


def logsoftmax(x, axis=-1):
    """Apply log softmax to a vector of logits x along an axis"""
    return x - logsumexp(x, axis, keepdims=True)


def fastvar(x, axis, keepdims):
    """A fast but less numerically-stable variance calculation than np.var."""
    return (
        np.mean(x ** 2, axis, keepdims=keepdims) - np.mean(x, axis, keepdims=keepdims) ** 2
    )
